# -*- coding: utf-8 -*-
# This file is part of ranger, the console file manager.
# License: GNU GPL version 3, see the file "AUTHORS" for details.
# Author: Abdó Roig-Maranges <abdo.roig@gmail.com>, 2011
#
# vcs - a python module to handle various version control systems

import os

from .vcs import VcsError, Vcs

# vim: expandtab:shiftwidth=4:tabstop=4:softtabstop=4:textwidth=80
